/* GPDOCXO
   Display status of GPS and Lock condition
   Monitor activity of Arduino processes, (visual indication of lock up (First heart tick)
   Second heart shows NMEA messages been received by unit been processed
   GPS Monitor physical condition of module with alert due to failure with auto boot
   UTC Clock with Time, Day
*/
// Libraries used
#include <TinyGPS++.h>
#include <AltSoftSerial.h>
#include <LiquidCrystal_I2C.h>  // I2C LCD Lib

// Defining associated library functions 
LiquidCrystal_I2C lcd(0x3F, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  // If I2C address is not known Use the supplied code to find out 0x3F
AltSoftSerial mySerial(8, 9); // Pin 8 RX, Pin 9 TX
TinyGPSPlus gps;

// System Timers
byte customChar[8] = {  0b00000, 0b01010, 0b11111, 0b11111, 0b01110, 0b00100, 0b00000, 0b00000 };
unsigned long time_now = 0;
unsigned long micro_now = 0;
const int period = 1000;   // Trim Tick period (1 Second)
const int MicroPeriod = 1; // Microseconds (Microtimer)
const int Baud = 9600;     //If you wish to push above this rate, match the GPS module to your selected BAUD
const int Interval = 10;   // Warm up time in Minutes (Settle Interval/2) The Count shows seconds remaining

// GPS obtained DATA variables
short GPSnum = 0;
short GPSlock = 0;
short GPSlat = 0;
short GPSlng = 0;
short ShowTime = 0;

// Logic Control variables
short TurnonTimer;
short WarmupDone = 0;
short Settle = 0;
short Pass = 0;
short Lock;
short Tick = 0;
short MicroTick;
short Blink;
short OKGPS = 0;
short  fail;
short ModuleFail;
short NoSettle = 0;

// Setup & Startup message, initialising LCD Panel and Serial Ports, 
void setup()
{
  lcd.begin(16, 2);
  lcd.clear();
  lcd.setCursor(3, 0);
  lcd.print(F("GPSD OCXO"));
  lcd.setCursor(0, 1);
  lcd.print(F("10 MHz Reference"));
  lcd.setCursor(0, 2);
  delay(3000);
  lcd.clear();
  lcd.createChar(0, customChar);
  lcd.setCursor(0, 0);
  lcd.print(F("Sat"));
  lcd.setCursor(13, 0);
  lcd.print(F("Loc"));
  Serial.begin(Baud);
  mySerial.begin(Baud);
}
// Called to soft reset and restart 
void(* resetFunc) (void) = 0;  //declare reset function @ address 0

void loop()
{
  getdata();      // Only get data when required
  Blink ++;       // Blinks heart for active CPU or NMEA Data
  OvenWarmUp();   // Call this only on Cold Start
  SysStatus();    // Test for normal conditions
  if (ShowTime == 1 && WarmupDone == 1 && Settle == 0) displayTime();  // If all conditions are normal display UTC Tine and Day
  LockDetect();   // Test that GPS is locked
  Settling();     // Settling timer after signal loss and reacquisition 
  SigLost();      // Signal Lost detector
  WatchDog();     // Processes monitor (display Heart Tick 
}

void getdata() // Read NMEA DATA or restart System on Faulty received DATA
{
  fail ++;  // Counts number of read attempts that fail 
  if (!mySerial.available())
  {
    if (fail >= 55)      // Test that the GPS Module is functioning >55 GPS system has failed 
    {
      ModuleFail = 1;    // GPS Fail  Warning - Requires cold start to reset
      OKGPS = 0;
      GPSnum = 0;
      GPSlock = 0;
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("GPS  Module Fail"));
      delay(3000);
      lcd.setCursor(3, 1);
      lcd.print(F("Restarting"));
      delay(1000);
      resetFunc();  //call reset
    }
  }

  while (mySerial.available()) // Valid NNEA Data received and processed
  {
    gps.encode(mySerial.read());
    GPSnum  = gps.satellites.value();
    GPSlat = gps.location.lat();
    GPSlng  = gps.location.lng();
    if (GPSlat != 0 && GPSlng != 0 && GPSnum > 0)
    {
      GPSlock = 1;
    }
    else GPSlock = 0;
    OKGPS = 1;      // NMEA message received Flag
    fail = 1;
  }
}

void SysStatus()    // System test that all is as it should be, evoking Settling time on Sat signal Loss
{
  if (Lock == 1 && Settle == 0 &&  WarmupDone == 1 && ShowTime == 0)
  {
    TickTimer(1);
    lcd.setCursor(0, 1);
    lcd.print(F("FR 10.000000 MHz"));
    delay(2000);
    if (Tick <= 5)  // After 5 seconds switch display to show Time and Day 
    {
      ShowTime = 1;
    }
    else
    {
      TickTimer(0);
      Tick = 0;
    }
  }
  if (Lock == 0 && GPSnum == 0 && WarmupDone == 1)  // Display Loss of PLL lock 
  {
    ShowTime = 0;
    if (Pass == 1)
    {
      lcd.setCursor(0, 1);
      lcd.print(F("PLL Missing Lock"));
    }
    Pass = 0;
  }
}

void OvenWarmUp()           // User alert - timer that OCXO is in the warm up cycle
{
  short Wt = Interval * 60; // Warmup Time defined by interval constant
  if (WarmupDone == 0 )     // This variable is set to zero at start-up indicating a cold start
  {
    TickTimer(1);
    lcd.setCursor(0, 2);
    lcd.print(F("OCXO LowTemp"));
    Count(Wt);
    if (Tick == Wt)         // Warm Up completed
    {
      TickTimer(0);
      WarmupDone = 1;
      Tick = 0;
      lcd.setCursor(0, 2);
      lcd.print(F("               "));
    }
  }
}

void LockDetect()          // Test that Valid Satellite Lock with valid DATA
{
  if (GPSnum > 0 && OKGPS == 1 && GPSlock == 1)  // Number of satellites
  {
    lcd.setCursor(3, 0);
    lcd.print(GPSnum);
    if (GPSnum <= 9)
    {
      lcd.setCursor(4, 0);
      lcd.print(F(" "));
    }
  }
  if (GPSnum == 0)
  {
    lcd.setCursor(3, 0);
    lcd.print(F("  "));
    Lock = 0;
  }
  if (GPSlock == 1)    // GPS Lock "Lock = 1"
  {
    lcd.setCursor(11, 0); lcd.print(F("  ")); // Lock
    Lock = 1;
  }
  if (Lock == 0)      // GPS No Lock
  {
    //Lock = 0 ;
    lcd.setCursor(3, 0);
    lcd.print(F("0 "));
    lcd.setCursor(11, 0);
    lcd.print(F("N-"));
  }
}

void SigLost() // Adjusting Logic to loss of signal, evoking Settling Timer on reacquisition
{
  if ( Lock == 0 && WarmupDone == 1 )    // Loss of lock after warm up "No Lock"
  {
    Pass = 1;
    Settle = 1;
    Tick = 0;
  }
}

void Settling() // Settling Timer after reacquisition
{
  short St = Interval * 24; //Settling Time 1/2 of warmup time
  if (GPSlock == 0 && Settle == 1 && Pass == 1)  // Exits Settling if Loss of lock
  {
    Tick = 0;
    Lock = 0;
    return ;
  }
  if (Lock == 1 && Settle == 1)
  {
    TickTimer(1);
    lcd.setCursor(0, 1);
    lcd.print(F("PLL Setlling "));
    Count(St);
    if (Tick == St)
    {
      TickTimer(0);
      Settle = 0;
      Tick = 0;
    }
  }
}

void microTimer(int TimerOn) // System 1 microsecond Timer
{
  short Start;
  short Stop;
  if (TimerOn == 1)
  {
    Start = 1;
    Stop = 0;
  }
  if (TimerOn == 0)
  {
    Start = 0;
    Stop = 1;
  }
  if (Start == 1)
  {
    if ((unsigned long)(micros() - micro_now) > MicroPeriod)
    {
      MicroTick ++;
      micro_now = micros();
    }
    if (Stop == 1) MicroTick = 0;
  }
}

void TickTimer(int TimerOn) // System 1 second Timer
{
  short Start;
  short Stop;
  if (TimerOn == 1)
  {
    Start = 1;
    Stop = 0;
  }
  if (TimerOn == 0)
  {
    Start = 0;
    Stop = 1;
  }
  if (Start == 1)
  {
    if ((unsigned long)(millis() - time_now) > period)
    {
      Tick ++;
      time_now = millis();
    }
    if (Stop == 1)
    {
      Tick = 0;
    }
  }
}

void WatchDog()    // Hearts toggling - Shows parts of the code operating as it should
{
  if (Blink == 10) // CPU Watch dog
  {
    lcd.setCursor(7, 0);
    lcd.write((byte)0);
  }
  if (Blink == 20)
  {
    lcd.setCursor(7, 0);
    lcd.print(F(" "));
  }
  if (Blink == 30)
  {
    Blink = 0;
  }

  if (OKGPS == 1 && Blink == 7)
  {
    lcd.setCursor(8, 0);
    lcd.write((byte)0);
  }

  if (OKGPS == 1 && Blink == 14)
  {
    lcd.setCursor(8, 0);
    lcd.print(F("  "));
  }

  if (Blink == 21)
  {
    Blink = 0;
  }
}

void Count(int x)  // Display countdown of Warmup or Settling timers
{
  short Counter;
  Counter = (x - Tick);
  if (Counter >= 100 )
  {
    lcd.setCursor(13, 1);
  }
  if (Counter <= 99)
  {
    lcd.setCursor(13, 1);
    lcd.print(F(" "));
    lcd.setCursor(14, 3);
  }
  if (Counter <= 9 )
  {
    lcd.setCursor(14, 1);
    lcd.print(F(" "));
    lcd.setCursor(15, 1);
  }
  if (Counter <= 0 )
  {
    lcd.setCursor(14, 1);
    lcd.print(F(" "));
  }
  if (Counter != 0)
  {
    lcd.print(Counter);
  }
  else lcd.setCursor(15, 1); lcd.print(F(" "));
}

void displayTime() // UTC & Date generation Routines
{
  int m;           // Month Entry
  int d;           // Day Entry
  int yy;          // Last 2 digits of the year (ie 2016 would be 16)
  int yyyy;        // Year Entry
  int c;           // Century (ie 2016 would be 20)
  int mTable;      // Month value based on calculation table
  int SummerDate;  // Add values combined in prep for Mod7 calc
  int DoW;         // Day of the week value (0-6)
  int leap;        // Leap Year or not
  int cTable;      // Century value based on calculation table
  Tick = 5;
  {
    lcd.setCursor(0, 1);
    lcd.print(F("UTC "));

    if (gps.time.hour() < 10) lcd.print(F("0"));
    lcd.print(gps.time.hour());
    lcd.print(F(":"));
    if (gps.time.minute() < 10) lcd.print(F("0"));
    lcd.print(gps.time.minute());
    lcd.print(F(":"));
    if (gps.time.second() < 10) lcd.print(F("0"));
    lcd.print(gps.time.second());
  }
  m = gps.date.month();
  d = gps.date.day();
  yyyy = gps.date.year();

  // Leap Year Calculation
  if ((fmod(yyyy, 4) == 0 && fmod(yyyy, 100) != 0) || (fmod(yyyy, 400) == 0))
  {
    leap = 1;
  }
  else
  {
    leap = 0;
  }
  // Limit results to year 1900-2299 (to save memory)
  while (yyyy > 2299)
  {
    yyyy = yyyy - 400;
  }
  while (yyyy < 1900)
  {
    yyyy = yyyy + 400;
  }
  // Calculating century
  c = yyyy / 100;
  // Calculating two digit year
  yy = fmod(yyyy, 100);
  // Century value based on Table
  if (c == 19) {
    cTable = 1;
  }
  if (c == 20) {
    cTable = 0;
  }
  if (c == 21) {
    cTable = 5;
  }
  if (c == 22) {
    cTable = 3;
  }
  // Jan and Feb calculations affected by leap years
  if (m == 1)
  { if (leap == 1) {
      mTable = 6;
    }
    else          {
      mTable = 0;
    }
  }
  if (m == 2)
  { if (leap == 1) {
      mTable = 2;
    }
    else          {
      mTable = 3;
    }
  }
  // Other months not affected and have set values
  if (m == 10) {
    mTable = 0;
  }
  if (m == 8) {
    mTable = 2;
  }
  if (m == 3 || m == 11) {
    mTable = 3;
  }
  if (m == 4 || m == 7) {
    mTable = 6;
  }
  if (m == 5) {
    mTable = 1;
  }
  if (m == 6) {
    mTable = 4;
  }
  if (m == 9 || m == 12) {
    mTable = 5;
  }
 
  // Enter the data into the formula
  SummerDate = d + mTable + yy + (yy / 4) + cTable;
  // Find remainder
  DoW = fmod(SummerDate, 7);
  // Remainder determines day of the week
  lcd.setCursor(13, 1);
  if (DoW == 0) {
    lcd.print(F("Sat "));
  }
  if (DoW == 1) {
    lcd.print(F("Sun "));
  }
  if (DoW == 2) {
    lcd.print(F("Mon "));
  }
  if (DoW == 3) {
    lcd.print(F("Tue "));
  }
  if (DoW == 4) {
    lcd.print(F("Wed "));
  }
  if (DoW == 5) {
    lcd.print(F("Thu "));
  }
  if (DoW == 6) {
    lcd.print(F("Fri "));
  }
}
